import inspect
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Callable, Dict, Type, TypeVar

from omegaconf import DictConfig, ListConfig, OmegaConf
from typing_extensions import dataclass_transform

from .traversal import traverse
from .types import get_type


class FlexyClass:
    ...


ClsToFlex = TypeVar('ClsToFlex')


@dataclass_transform()
def flexyclass(cls: Type[ClsToFlex]) -> Type[ClsToFlex]:
    """A flexyclass is like a dataclass, but it supports partial
    specification of properties."""

    if not inspect.isclass(cls):
        raise TypeError(f'flexyclass must decorate a class, not {cls}')

    new_cls = type(f'FlexyClass{cls.__name__}',
                   (dataclass(cls), FlexyClass),
                   {})

    return new_cls


def flexyfactory(
    flexy_cls: Type[Dict[str, Any]],
    **kwargs: Any
) -> Dict[str, Any]:
    def factory_fn() -> Dict[str, Any]:
        return flexy_cls(**kwargs)
    return field(default_factory=factory_fn)


DictOrListConfig = TypeVar('DictOrListConfig', DictConfig, ListConfig)


def unlock_all_flexyclasses(
    cast_fn: Callable[..., DictOrListConfig]
) -> Callable[..., DictOrListConfig]:

    @wraps(cast_fn)
    def unlock_fn(*args, **kwargs):
        config_node = cast_fn(*args, **kwargs)

        for spec in traverse(node=config_node,
                             include_nodes=True,
                             include_root=True):

            if not isinstance(spec.value, DictConfig):
                # only DictConfigs can be flexyclasses
                continue

            typ_ = get_type(spec.value)
            if typ_ and issubclass(typ_, FlexyClass):
                OmegaConf.set_struct(spec.value, False)

        return config_node

    return unlock_fn



from dataclasses import field, dataclass
from omegaconf import II, SI, MISSING

from .commandline import cli
from .core import (
    cast,
    from_dataclass,
    from_dict,
    from_file,
    from_none,
    from_options,
    from_string,
    merge,
    to_dict,
    to_yaml,
    validate,
)
from .traversal import traverse
from .types import get_type
from .initialize import Target, init
from .logging import configure_logging
from .resolvers import register, all_resolvers
from .flexyclasses import flexyclass, flexyfactory


__all__ = [
    'all_resolvers',
    'cast',
    'cli',
    'configure_logging',
    'dataclass',
    'field',
    'flexyclass',
    'flexyfactory',
    'from_dataclass',
    'from_dict',
    'from_file',
    'from_none',
    'from_options',
    'from_string',
    'get_type',
    'II',
    'init',
    'merge',
    'MISSING',
    'register',
    'SI',
    'Target',
    'to_dict',
    'to_yaml',
    'traverse',
    'validate',
]



import hashlib
import pickle
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Optional, Sequence, Union

from .initialize import Target


def cache_to_disk(kwargs: Optional[Sequence[str]] = None,
                  location: Optional[Union[str, Path]] = None) -> Callable:
    """Decorator to cache a function's output to disk.

    Args:
        kwargs: The name of the values in the function's signature to
            use to determine the filename of the cache file.
            If None, all values are used.
        location: The directory where the cache file will be stored.
            If None, ~/.cache is used.

    Returns:
        A decorator that enables caches the output of a function
            decorated with it.
    """

    if location is None:
        location = (Path('~') / '.cache').expanduser().absolute()
    path_location = Path(location)

    # make caching directory if it doesn't exist
    if not path_location.exists():
        path_location.mkdir(parents=True)

    def decorator(func: Callable,
                  kwargs_to_cache: Optional[Sequence[str]] = kwargs,
                  location: Path = path_location) -> Callable:
        func_name = Target.to_string(func)

        @wraps(func)
        def wrapper(
            *args: Any,
            __kwargs_to_cache__: Optional[Sequence[str]] = kwargs_to_cache,
            __location__: Path = location,
            __invalidate__: bool = False,
            __func_name__: str = func_name,
            **kwargs: Any
        ) -> Any:
            if args:
                msg = 'You cannot pass non-positional arguments when caching'
                raise ValueError(msg)

            # always cache in the same order
            if __kwargs_to_cache__ is None:
                __kwargs_to_cache__ = tuple(kwargs.keys())
            __kwargs_to_cache__ = sorted(__kwargs_to_cache__)

            # collect hash here
            h = hashlib.sha1()

            # hash function name
            h.update(__func_name__.encode('utf-8'))

            # hash all kwargs
            for k, v in kwargs.items():
                if k in __kwargs_to_cache__:
                    # include name of key in cache name
                    h.update(pickle.dumps((k, v)))

            # digest and give .pickle extension
            cache_path = __location__ / f'{h.hexdigest()}.pickle'

            if not cache_path.exists() or __invalidate__:
                # cache miss
                resp = func(**kwargs)
                with open(cache_path, 'wb') as f:
                    pickle.dump(resp, f)
            else:
                # cache hit
                with open(cache_path, 'rb') as f:
                    resp = pickle.load(f)

            # return whatever the function returns
            return resp

        return wrapper

    return decorator



import functools
import importlib
import importlib.util
import inspect
import itertools
from typing import Any, Callable, Optional, Type, TypeVar

from omegaconf import DictConfig

from .core import ConfigType, cast
from .utils import clean_multiline


class InitLater(functools.partial):

    # inherits slots from functools.partial
    __slots__ = "type_",

    def __new__(cls, func, type_: Optional[type] = None, /, *args, **keywords):
        cl = super().__new__(cls, func, *args, **keywords)
        cl.type_ = type_
        return cl

    def get_kw(self: 'InitLater', *args, **kwargs):
        """Shortcut for accessing parameters that have been
        provided to an InitLater object"""
        return self.keywords.get(*args, **kwargs)

    def pop_kw(self: 'InitLater', *args, **kwargs):
        """Shortcut for popping parameters that have been
        provided to an InitLater object"""
        return self.keywords.pop(*args, **kwargs)

    @staticmethod
    def _no_op():
        ...

    @classmethod
    def no_op(cls: Type['InitLater']) -> 'InitLater':
        """Create an init later that does nothing.
        Useful for when trying to instantiate from None."""
        return cls(cls._no_op)

    def __bool__(self: 'InitLater') -> bool:
        # this allows us to check if the InitLater object
        # is "True" (that is, it is a real wrapped function),
        # or if it is "False" (it's a no-op)
        return self.func != self._no_op

    def __call__(self: 'InitLater', /, *args, **kwargs):
        # recursively call deferred initialization if
        # we encounter another InitLater
        args = tuple(v() if isinstance(v, InitLater) else v
                     for v in itertools.chain(self.args, args))
        kwargs = {k: v() if isinstance(v, InitLater) else v
                  for k, v in {**self.keywords, **kwargs}.items()}
        try:
            out = self.func(*args, **kwargs)

            # if type is provided, then we check if the object that
            # has been initialized here is of the expected type.
            # note that this only works for top-level init, and
            # does not recursively check.

            try:
                type_check_out = isinstance(out, self.type_)
            except TypeError:
                # Instance check fails if we can't check for type for
                # some reason, e.g. if self.type_ is none, self.type_ is a
                # Protocol or self._type is an instance of a class.
                #
                # If the instance check fails, we assume that the type check
                # went through, and return the object
                type_check_out = True

            if not type_check_out:
                msg = (f"Initialized object `{out}` is not the right type: "
                       f"expected `{self.type_}`, got {type(out)}")
                raise TypeError(msg)

            return out
        except Exception as e:
            ex_name = type(e).__name__
            fn_name = repr(self.func)
            msg = clean_multiline(f'''
                An error occurred while trying to initialize {fn_name}
                with arguments "{args}" and kwargs "{kwargs}":
                {ex_name}("{' '.join(map(str, e.args))}")
            ''')
            raise RuntimeError(msg).with_traceback(e.__traceback__)


class Target:
    @staticmethod
    def _is_module(path: str) -> bool:
        try:
            spec = importlib.util.find_spec(path)
            return spec is not None
        except ModuleNotFoundError:
            return False

    @classmethod
    def to_string(cls: Type['Target'], callable: Any) -> str:
        if inspect.isclass(callable):
            return f'{callable.__module__}.{callable.__name__}'
        elif inspect.ismethod(callable):
            return f'{cls.to_string(callable.__self__)}.{callable.__name__}'
        elif inspect.isfunction(callable):
            return f'{callable.__module__}.{callable.__name__}'
        else:
            raise TypeError(f'`{callable}` is not a callable')

    @classmethod
    def from_string(cls: Type['Target'], path: str) -> Any:
        if cls._is_module(path):
            return importlib.import_module(path)
        elif '.' in path:
            m_name, c_name = path.rsplit('.', 1)
            if cls._is_module(m_name):
                container = importlib.import_module(m_name)
            else:
                container = cls.from_string(m_name)
            callable_ = getattr(container, c_name, None)
        else:
            callable_ = globals().get(path, getattr(__builtins__, path, None))

        if callable_ is None:
            raise ImportError(f'Cannot find callable at {path}')

        return callable_


InitT = TypeVar('InitT')
CallableT = TypeVar('CallableT', bound=Callable)


class init:
    TARGET: str = '_target_'

    @classmethod
    def callable(cls: Type['init'],
                 config: Optional[ConfigType] = None,
                 target: Optional[str] = None,
                 _type_: Optional[CallableT] = None) -> CallableT:

        if target is not None and config is not None:
            raise ValueError('Cannot specify both `target` and `config`')
        elif target is None:
            config = cast(config)
            try:
                target = str(config[cls.TARGET])
            except KeyError:
                raise KeyError(f'Config `{config}` has no `{cls.TARGET}` key!')

        return Target.from_string(target)

    @classmethod
    def later(
        cls: Type['init'],
        config: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        **kwargs: Any
    ) -> Callable[..., InitT]:
        """Return a InitLater object to be used to instantiate a
        new class or call a function"""

        # if no config is provided, we return a function
        if config is None:
            return InitLater.no_op()

        # ensure we are working with a ConfigDict
        config_node = cast(config)

        # try to get the target callable from either the config
        # or keyword arguments passed to the init function
        if cls.TARGET in config_node:
            fn = cls.callable(config=config_node)
        elif cls.TARGET in kwargs:
            fn = cls.callable(target=kwargs[cls.TARGET])
        else:
            raise ValueError(f'Cannot instantiate from `{config_node}` and '
                             f'`{kwargs}`: `{cls.TARGET}` keyword missing')

        def _recursive_init(param):
            must_recursive_init = (
                _recursive_ and
                isinstance(param, (DictConfig, dict)) and
                cls.TARGET in param
            )
            if must_recursive_init:
                param = cls.later(config=param, _recursive_=True)
            return param

        init_call_dict = {str(k): _recursive_init(v)
                          for k, v in {**config_node, **kwargs}.items()
                          if k != cls.TARGET}

        return InitLater(fn, _type_, **init_call_dict)

    @classmethod
    def now(
        cls: Type['init'],
        config: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        **kwargs: Any
    ) -> InitT:
        """Create a later, but initialize it now!"""
        init_call = cls.later(config=config,
                              _type_=_type_,
                              _recursive_=_recursive_,
                              **kwargs)
        return init_call()

    def __new__(
        cls: Type['init'],
        config: Optional[Any] = None,
        _type_: Optional[Type[InitT]] = None,
        _recursive_: bool = True,
        **kwargs: Any
    ) -> InitT:
        """Alias init(...) to init.now(...)"""
        return cls.now(config=config,
                       _type_=_type_,
                       _recursive_=_recursive_,
                       **kwargs)



import os
from argparse import ArgumentParser
from enum import Enum
from functools import partial
from inspect import getfile, getfullargspec, isclass
from pathlib import Path
from typing import Any, Callable, Optional, Protocol, Sequence, Type, overload

from omegaconf import DictConfig

from .core import (_DataClass, from_dataclass, from_file, from_options, merge,
                   traverse, validate, from_none)
from .initialize import InitLater
from .utils import PrintUtils, clean_multiline


class CliFlags(Enum):
    CONFIG = 'config'
    PARSED = 'parsed'
    QUIET = 'quiet'
    INPUTS = 'inputs'
    OPTIONS = 'options'
    DEBUG = 'debug'


def make_flags(opt_name: CliFlags) -> Sequence[str]:
    """Simple helper to create a list of flags for an
    option based on its name"""
    return f'-{opt_name.value[0]}', f'--{opt_name.value}'


class MainFnProtocol(Protocol):
    __name__: str

    @overload
    def __call__(self, config: Any) -> Any:
        ...

    @overload
    def __call__(self, config: Any, *args: Any) -> Any:
        ...

    def __call__(self, config: Any, *args: Any, **kwargs: Any) -> Any:
        ...


class PrintFnProtocol(Protocol):
    def __call__(self, txt: str, /, *args: Any, **kwargs: Any) -> None:
        ...


class DecoratedMainFnProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...


def check_if_callable_can_be_decorated(func: MainFnProtocol):
    expected_args = getfullargspec(func).args
    if len(expected_args) == 0:
        msg = (f'Function `{func.__name__}` cannot be decorated '
               f'by `config_to_program` because it does '
               f'not accept any argument.')
        raise RuntimeError(msg)
    elif len(expected_args) > 1:
        msg = (f'Function `{func.__name__}` cannot be decorated by '
               f' `config_to_program` because it expects '
               f'{len(expected_args)} > 1; If you want to pass extra '
               f'arguments to this function, use kwargs with default values.')
        raise RuntimeError(msg)


def check_if_valid_main_args(func: MainFnProtocol, args: Sequence[Any]):
    if len(args):
        msg = (f'After decorating `{func.__name__}` with '
               f'`config_to_program`, do not provide any additional '
               f'arguments while invoking it; any additional parameter '
               f'should be passed as a keyword argument.')
        raise RuntimeError(msg)


def make_cli_argument_parser(func: MainFnProtocol,
                             name: str) -> ArgumentParser:
    """Sets up argument parser ahead of running the CLI. This includes
    creating a help message, and adding a series of flags."""

    # we find the path to the script we are decorating with the
    # cli so that we can display that to the user.
    current_dir = Path(os.getcwd())
    path_to_fn_file = Path(getfile(func))
    rel_fn_file_path = str(path_to_fn_file).replace(str(current_dir), '')

    # Program name and usage printed here.
    prog = f'Parser for configuration {name}'
    usage = clean_multiline(f'''
        python3 {rel_fn_file_path} '
        {{{"/".join(make_flags(CliFlags.CONFIG))} config_file.yaml}}
        {{{"/".join(make_flags(CliFlags.OPTIONS))}}}
        {{{"/".join(make_flags(CliFlags.INPUTS))}}}
        {{{"/".join(make_flags(CliFlags.DEBUG))}}}
        {{{"/".join(make_flags(CliFlags.PARSED))}}}
        {{{"/".join(make_flags(CliFlags.QUIET))}}}
        param1=value1, …, paramN=valueN'
    ''')
    ap = ArgumentParser(prog=prog, usage=usage)

    # add options
    msg = 'A path to a YAML file containing a configuration.'
    ap.add_argument(*make_flags(CliFlags.CONFIG),
                    default=[],
                    help=msg,
                    nargs='*',
                    metavar='/path/to/config.yaml')

    msg = 'Print all default options and CLI flags.'
    ap.add_argument(*make_flags(CliFlags.OPTIONS),
                    action='store_true',
                    help=msg)

    msg = 'Print the input configuration.'
    ap.add_argument(*make_flags(CliFlags.INPUTS),
                    action='store_true',
                    help=msg)

    msg = 'Print the parsed configuration.'
    ap.add_argument(*make_flags(CliFlags.PARSED),
                    action='store_true',
                    help=msg)

    msg = 'Enter debug mode by setting global logging to DEBUG.'
    ap.add_argument(*make_flags(CliFlags.DEBUG),
                    action='store_true',
                    help=msg)

    msg = 'If provided, it does not print the configuration when running.'
    ap.add_argument(*make_flags(CliFlags.QUIET),
                    action='store_true',
                    help=msg)
    return ap


def wrap_main_method(
    func: MainFnProtocol,
    name: str,
    config_node: DictConfig,
    print_fn: Optional[Callable] = None,
    *args: Any,
    **kwargs: Any
) -> Any:

    if not isinstance(config_node, DictConfig):
        raise TypeError("Config node must be a DictConfig")

    # Making sure I can decorate this function
    check_if_callable_can_be_decorated(func=func)
    check_if_valid_main_args(func=func, args=args)

    # Get argument parser and arguments
    ap = make_cli_argument_parser(func=func, name=name)
    opts, leftover_args = ap.parse_known_args()

    # set some default options for when no options are provided
    # printing_steps = PrintingSteps(opts)

    # setup debug
    if opts.debug:
        # relative import here not to mess things up
        from .logging import configure_logging
        configure_logging.debug()

    # Setup an utility to deal with printing
    pu = PrintUtils(print_fn=print_fn)

    # We don't run the main program if the user
    # has requested to print the any of the config.
    do_no_run = (opts.options or opts.inputs or opts.parsed)

    # Print default options if requested py the user
    if opts.options:
        params = traverse(config_node)

        cli_opts_repr = ('OPTS/CLI FLAG:', ) + tuple(
            f'{p.path} = ' + (str(p.value) if p.value != '' else "''")
            for p in params
        )
        pu.print(*cli_opts_repr, level_up=1)

    # load options from one or more config files;
    # if multiple config files are provided,
    # the latter ones can override the former ones.
    file_config = from_none()
    for config_file in opts.config:
        file_config = merge(file_config, from_file(config_file))

    # load options from cli
    cli_config = from_options(leftover_args)

    # merge file and cli config; cli config overrides file config
    input_config = merge(file_config, cli_config)

    # print both configs if requested
    if opts.inputs:
        pu.print('INPUT/CLI ARGS:', cli_config)
        pu.print('INPUT/CFG FILE:', file_config)

    if do_no_run and not opts.parsed:
        # if the user hasn't requested to print the parsed config
        # and we are not running the main program, we can exit here.
        return InitLater.no_op()

    # load configuration with node parsers
    parsed_config = merge(config_node, input_config)

    # check if all parameters are provided/resolved
    parsed_config = validate(parsed_config)

    # print it if requested
    if not(opts.quiet) or opts.parsed:
        pu.print('PARSE/ALL CFG:', parsed_config)

    if do_no_run:
        # we are not running because the user has requested to print
        # either the options, inputs, or parsed config.
        return InitLater.no_op()
    else:
        # we execute the main method and pass the parsed config to it
        return func(parsed_config, **kwargs)


def cli(
    config_node_cls: Optional[Type[Any]] = None,
    print_fn: Optional[PrintFnProtocol] = None
) -> Callable[[MainFnProtocol], DecoratedMainFnProtocol]:

    if config_node_cls is None:
        config_node = from_none()
        name = '<unnamed>'

    elif not(isclass(config_node_cls) and
             issubclass(config_node_cls, _DataClass)):
        msg = '`config_node` must be be decorated as a dataclass'
        raise ValueError(msg)
    else:
        config_node = from_dataclass(config_node_cls)
        name = config_node_cls.__name__

    def wrapper(func: MainFnProtocol) -> DecoratedMainFnProtocol:
        out = partial(wrap_main_method,
                      func=func,
                      name=name,
                      config_node=config_node,
                      print_fn=print_fn)
        return out

    return wrapper



from dataclasses import dataclass
from typing import Any, Iterator, Optional, Union

from omegaconf import MISSING, DictConfig, ListConfig, OmegaConf
from omegaconf.basecontainer import BaseContainer


@dataclass
class ParamSpec:
    key: Union[str, int, None]
    path: str
    value: Any
    parent: Optional[Union[DictConfig, ListConfig]]

    def is_node(self) -> bool:
        return isinstance(self.value, (DictConfig, ListConfig))


def traverse(
    node: BaseContainer,
    include_nodes: bool = False,
    include_root: bool = False
) -> Iterator[ParamSpec]:

    if include_root:
        yield ParamSpec(key=None, path='', value=node, parent=None)

    if isinstance(node, ListConfig):
        for i, element in enumerate(node):
            if isinstance(element, BaseContainer):
                for spec in traverse(element, include_nodes=include_nodes):
                    yield ParamSpec(key=spec.key,
                                    path=f'[{i}].{spec.path}',
                                    value=spec.value,
                                    parent=spec.parent)

            if include_nodes or not isinstance(element, BaseContainer):
                yield ParamSpec(key=i,
                                path=f'[{i}]',
                                value=element,
                                parent=node)

    elif isinstance(node, DictConfig):
        for key in node.keys():
            if OmegaConf.is_missing(node, key):
                value = MISSING
            elif OmegaConf.is_interpolation(node, str(key)):
                # We don't want to resolve interpolations for now,
                # as they might not be resolvable at the moment
                value = OmegaConf.to_container(node)[key]  # type: ignore
            else:
                value = OmegaConf.select(node, str(key))

            if isinstance(value, BaseContainer):
                for spec in traverse(value, include_nodes=include_nodes):
                    yield ParamSpec(key=spec.key,
                                    path=f'{key}.{spec.path}',
                                    value=spec.value,
                                    parent=spec.parent)

            if include_nodes or not isinstance(value, BaseContainer):
                yield ParamSpec(key=str(key),
                                path=str(key),
                                value=value,
                                parent=node)

    else:
        raise ValueError(f'Cannot traverse `{node}`; DictConfig or ListConfig '
                         f'expected, but got `{type(node)}`.')
